/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.framework.core;

import java.util.*;

/**
 * Corresponds to the &lt;SyncBody&gt; element in the SyncML represent DTD
 *
 *  @author Stefano Fornari @ Funambol
 *  @see    SyncHdr
 *
 *  @version $Id: SyncBody.java,v 1.5 2005/03/02 20:57:37 harrie Exp $
 */
public final class SyncBody
implements java.io.Serializable {
    // ------------------------------------------------------------ Private data
    private ArrayList commands = new ArrayList();
    private Boolean finalMsg;

    // ------------------------------------------------------------ Constructors

    /** For serialization purposes */
    protected SyncBody() {}

    /**
     * Create a new SyncBody object. The commands in <i>commands</i>
     * must be of the allowed types.
     *
     * @param commands The array elements must be an instance of one of these
     *                 classes: {@link Alert},  {@link Atomic}, {@link Copy},
     *                 {@link Exec}, {@link Get}, {@link Map}, {@link Put},
     *                 {@link Results}, {@link Search}, {@link Sequence},
     *                 {@link Status}, {@link Sync}
     * @param finalMsg is true if this is the final message that is being sent
     *
     */
    public SyncBody( final AbstractCommand[] commands, final boolean finalMsg) {

        setCommands(commands);
        this.finalMsg = (finalMsg) ? new Boolean(finalMsg) : null;
    }

    // ---------------------------------------------------------- Public methods

    /**
     *
     *  @return the return value is guaranteed to be non-null. Also,
     *          the elements of the array are guaranteed to be non-null.
     *
     */
    public ArrayList getCommands() {
        return this.commands;
    }


    /**
     * This get method always returns empty ArrayList. This is used
     * in the JiBX mapping in order to unmarshall commands that could
     * be followed by the tag &lt;Final/&gt;
     *
     * @return empty ArrayList
     */
    public ArrayList getEmptyCommands() {
        return new ArrayList();
    }

    /**
     * Add the list of commands to the global ArrayList of commands.
     * This add method was introduced because into SyncML message
     * the tag &lt;Final/&gt; could be inserted between the AbstractCommand
     * and not at the end of SyncBody.
     *
     * The given commands must be of the allowed types.
     *
     * @param commands the commands - NOT NULL and o the allowed types
     *
     * @throws IllegalArgumentException if the constraints are not met
     */
    public void addAllCommands(List commands) {
		if (commands == null) {
			return;
		}
        for (int i = 0; i < commands.size(); i++) {
            if (commands.get(i) == null) {
                throw new IllegalArgumentException("commands[" + i +"] cannot be null");
            } else if (   (!(commands.get(i) instanceof Alert))
                       && (!(commands.get(i) instanceof Add))
                       && (!(commands.get(i) instanceof Atomic))
                       && (!(commands.get(i) instanceof Copy))
                       && (!(commands.get(i) instanceof Delete))
                       && (!(commands.get(i) instanceof Exec))
                       && (!(commands.get(i) instanceof Get))
                       && (!(commands.get(i) instanceof Map))
                       && (!(commands.get(i) instanceof Put))
                       && (!(commands.get(i) instanceof Replace))
                       && (!(commands.get(i) instanceof Results))
                       && (!(commands.get(i) instanceof Search))
                       && (!(commands.get(i) instanceof Sequence))
                       && (!(commands.get(i) instanceof Status))
                       && (!(commands.get(i) instanceof Sync))) {
                throw new IllegalArgumentException(
                    "commands[" + i + "] cannot be a " + ((AbstractCommand)commands.get(i)).getName()
                );
            }
        }

        this.commands.addAll(commands);
    }

    /**
     * Sets the sequenced commands. The given commands must be of the allowed
     * types.
     *
     * @param commands the commands - NOT NULL and o the allowed types
     *
     * @throws IllegalArgumentException if the constraints are not met
     */
    public void setCommands(AbstractCommand[] commands) {
        if (commands == null) {
            throw new IllegalArgumentException("commands cannot be null");
        }
        this.commands.clear();
        addAllCommands(Arrays.asList(commands));
    }

    /**
     * Sets the message as final
     *
     * @param finalMsg the Boolean value of finalMsg property
     */
    public void setFinalMsg(Boolean finalMsg) {
        this.finalMsg = (finalMsg.booleanValue()) ? finalMsg : null;
    }

    /**
     * Gets the value of finalMsg property
     *
     * @return true if this is the final message being sent, otherwise false
     *
     */
    public boolean isFinalMsg() {
        return (finalMsg != null);
    }

    /**
     * Gets the value of finalMsg property
     *
     * @return true if this is the final message being sent, otherwise null
     *
     */
    public Boolean getFinalMsg() {
        if (!finalMsg.booleanValue()) {
            return null;
        }
        return finalMsg;
    }
}