/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.engine;

import java.io.*;
import java.util.Map;
import java.util.logging.Logger;
import java.util.logging.Level;
import javax.naming.*;

import sync4j.framework.server.SyncResponse;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.config.ConfigurationConstants;
import sync4j.framework.config.ConfigClassLoader;
import sync4j.framework.config.ConfigurationException;
import sync4j.framework.core.Constants;
import sync4j.framework.core.StatusCode;
import sync4j.framework.core.SyncML;
import sync4j.framework.core.Sync4jException;
import sync4j.framework.core.Util;
import sync4j.framework.engine.MessageSizeCalculator;
import sync4j.framework.protocol.ProtocolException;
import sync4j.framework.protocol.ProtocolUtil;
import sync4j.framework.server.session.*;
import sync4j.framework.server.error.*;
import sync4j.framework.tools.WBXMLTools;
import sync4j.framework.tools.WBXMLSizeCalculator;
import sync4j.framework.tools.XMLSizeCalculator;
import sync4j.framework.tools.beans.*;

import sync4j.framework.engine.pipeline.PipelineManager;
import sync4j.framework.engine.pipeline.MessageProcessingContext;

import sync4j.server.SyncMLCanonizer;
import sync4j.server.config.Configuration;
import sync4j.server.session.SyncSessionHandler;

import org.jibx.runtime.*;
import org.jibx.runtime.impl.*;

/**
 *  This class handles a synchronization request.
 *  <p>
 *  This server accepts synchronization requests addressed to the hostname
 *  indicated by the configuration property pointed by {CFG_SERVER_URI} (see
 *  Sync4j.xml).
 *  <p>
 *
 *  LOG NAME: sync4j.server
 *
 *  @author Luigia Fassina @ Funambol
 *
 *  @version $Id: SyncAdapter.java,v 1.32 2005/06/01 11:19:04 harrie Exp $
 *
 */
public class SyncAdapter
implements ConfigurationConstants {

    // ------------------------------------------------------- Private constants
    protected static final String CONFIG_SYNCML_CANONIZER
    = "sync4j/server/SyncMLCanonizer.xml";

    protected static final String PARAM_SESSION_ID = "sid";

    protected static final String HEADER_CONTENT_TYPE = "content-type";

    // ------------------------------------------------------------ Private data

    protected transient Logger log                = null;

    protected SessionHandler sessionHandler       = null;

    protected Configuration config                = null;

    protected PipelineManager pipelineManager     = null;

    protected MessageProcessingContext mpc        = null;

    protected SyncMLCanonizer syncMLCanonizer     = null;


    // ------------------------------------------------------------ Constructors

    /**
     * Creates a new SyncAdapter.
     *
     * @param config the configuration object
     */
    public SyncAdapter(Configuration config) {
        super();

        log = Sync4jLogger.getLogger("server");

        this.config = config;

        sessionHandler  = config.getSessionHandler();
        pipelineManager = config.getPipelineManager();

        try {
            syncMLCanonizer = (SyncMLCanonizer)config.getBeanInstanceByName(CONFIG_SYNCML_CANONIZER);
        } catch (Exception e) {
            log.throwing("SyncAdapter", "constructor", e);
            new Sync4jException("Error "
                                + e.getClass().getName()
                                + " creating the syncMLCanonizer: "
                                + e.getMessage()
                                ).printStackTrace();
        }

        mpc = new MessageProcessingContext();
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Tells the SyncAdapter that a new Synchronizatin session is starting
     *
     * @param sessionId the session id
     */
    public void beginSync(String sessionId) {
        mpc.setSessionProperty(mpc.PROPERTY_SESSION_ID, sessionId);
    }

    /**
     * Finalizes the sync session
     */
    public void endSync() {
        int currentState = sessionHandler.getCurrentState();

        switch (currentState) {
            case SessionHandler.STATE_END:
                sessionHandler.commit();
                break;

            case SessionHandler.STATE_ERROR:
                sessionHandler.abort(StatusCode.PROCESSING_ERROR);
                break;

            default:
                sessionHandler.abort(StatusCode.SESSION_EXPIRED);
                break;
        }
        log = null;
    }

    /**
     * process the incoming XML sync message
     *
     * @param msg must be non-null
     * @param parameters request parameters (at transport level)
     * @param headers request headers (at transport level)
     *
     */
    public SyncResponse processXMLMessage( final byte[] msg       ,
                                           final Map    parameters,
                                           final Map    headers   )
    throws ServerException {
        SyncResponse response = null;
        String inMessage = null;
        SyncML syncMLIn = null, syncMLOut = null;
        String charset = null;

        // We check first the prolog for the encoding.
        charset = getCharSetOfProlog(msg);
        if (charset == null) { // If not in the prolog, use mime-type.
            charset = getCharSet(headers);
            if (charset == null) { // If still not defined default to UTF-8.
                charset = "UTF-8";
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Charset (default): " + charset);
                }
            } else { // charset was defined in mime-type.
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Charset from mime-type: " + charset);
                }
            }
        } else { // Charset was defined in the prolog.
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Charset of prolog: " + charset);
            }
        }

        try {
            inMessage = new String(msg, charset);
        } catch (UnsupportedEncodingException e) {
            // if we have a not known encoding try the default.
            // FIXME, maybe better to bail out.
            inMessage = new String(msg);
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Message to translate into the SyncML object:\n" + inMessage);
        }

        inMessage = syncMLCanonizer.canonizeInput(inMessage);

        syncMLIn = convert(inMessage);

        mpc.setRequestProperty(mpc.PROPERTY_REQUEST_PARAMETERS, parameters);
        mpc.setRequestProperty(mpc.PROPERTY_REQUEST_HEADERS   , headers   );

        if (log.isLoggable(Level.FINE)) {
            log.fine("Calling input pipeline");
        }

        pipelineManager.preProcessMessage(mpc, syncMLIn);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("About processing message: " + Util.toXML(syncMLIn));
        }

        syncMLOut = processInputMessage(syncMLIn);

        setRespURI(syncMLOut, (String)mpc.getSessionProperty(mpc.PROPERTY_SESSION_ID));
        checkRespURISize(syncMLOut.getSyncHdr().getRespURI());

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Calling output pipeline");
        }

        pipelineManager.postProcessMessage(mpc, syncMLOut);

        byte[] out = convert(syncMLOut, charset);

        response = new Sync4jResponse(out, syncMLOut);

        return response;
    }

    /**
     * process the incoming WBXML sync message
     *
     * @param msg must be non-null
     * @param parameters request parameters (at transport level)
     * @param headers request headers (at transport level)
     *
     */
    public SyncResponse processWBXMLMessage( final byte[] msg       ,
                                             final Map    parameters,
                                             final Map    headers   )
    throws ServerException {
        SyncResponse response = null;
        String inMessage = null;
        SyncML syncMLIn = null, syncMLOut = null;
        String charset =  getCharSet(headers);

        //
        // Convert WBXML to XML, and then pass to processXMLMessage
        //
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Convertint message from wbxml to xml");
            log.finest("Char-set: " + charset + " (null, means wbxml defined)");
        }

        try {
            inMessage = WBXMLTools.wbxmlToXml(msg, charset);
        } catch (Sync4jException e) {
            throw new ServerException(e);
        }


        if (log.isLoggable(Level.FINEST)) {
            log.finest("Message to translate into the SyncML object:\n" + inMessage);
        }

        inMessage = syncMLCanonizer.canonizeInput(inMessage);

        syncMLIn = convert(inMessage);

        mpc.setRequestProperty(mpc.PROPERTY_REQUEST_PARAMETERS, parameters);
        mpc.setRequestProperty(mpc.PROPERTY_REQUEST_HEADERS   , headers   );

        if (log.isLoggable(Level.FINE)) {
            log.fine("Calling input pipeline");
        }
        pipelineManager.preProcessMessage(mpc, syncMLIn);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("About processing message: " + Util.toXML(syncMLIn));
        }

        syncMLOut = processInputMessage(syncMLIn);

        setRespURI(syncMLOut, (String)mpc.getSessionProperty(mpc.PROPERTY_SESSION_ID));
        checkRespURISize(syncMLOut.getSyncHdr().getRespURI());

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Calling output pipeline");
        }

        pipelineManager.postProcessMessage(mpc, syncMLOut);

        byte[] out = null;
        try {
            //
            // The marshalling is doing directly into WBXMLTools
            // because that method is calling in the other code too
            //
            out = WBXMLTools.toWBXML(syncMLOut);

        } catch(Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error processing the WBXML message: " + e.getMessage());
            }
            log.throwing("SyncAdapter", "processWBXMLMessage", e);

            throw new ServerException(e);
        }

        response = new Sync4jResponse(out, syncMLOut);


        return response;
    }

    /**
     * Returns the charset encoding if it was found in the prolog.
     * @param msg the message in which the prolog is searched.
     * @returns the charset if specified otherwise null.
     */
    private String getCharSetOfProlog(byte[] msg) {
        // Check if we have a start prolog, "<?xml".
        if ((msg[0] == '<') && (msg[1] == '?') && (msg[2] == 'x') && (msg[2] == 'm') && (msg[4] == 'l')) {
            // Now start looking for the end of the prolog.
            int length = msg.length - 1; // substract one for the double byte check.
            for (int i = 5; i < length; i++) {
                 if ((msg[i] == '?') && (msg[i+1] == '>')) {
                     // We found the end of the prolog.
                     String prolog = new String(msg, 0, i+1);
                     if (log.isLoggable(Level.FINEST)) {
                         log.finest("Message prolog: " + prolog);
                     }
                     // Search for encoding keyword.
                     int p = prolog.indexOf("encoding");
                     // First double quote after the encoding.
                     int q1 = prolog.indexOf('"', p);
                     // Second double quote after the encoding.
                     int q2 = prolog.indexOf('"', q1);
                     return prolog.substring(q1, q2);
                 }
            }
        }
        return null; // Was not defined
    }

    /**
     * Returns the charset encoding based on the header information.
     * @param headers the collection of headers of the request.
     * @returns the charset if specified otherwise null.
     */
    private String getCharSet(Map headers) {
        String contentType = (String)headers.get(HEADER_CONTENT_TYPE);
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Message has content type: " + contentType);
        }
        if (contentType != null) {
            // Check if we have a charset at all.
            int pointer = contentType.indexOf(';');
            if (pointer > 0) {
                // Now strip of the 'charset=' by searching of the first '='.
                // Only one is allowed anyway.
                pointer = contentType.indexOf('=', pointer);
                if (pointer > 0) {
                    String result = contentType.substring(pointer+1);
                    // Just trim whitespace incase the client is not polite.
                    return result.trim();
                }
            }
        }
        return null; // Was not defined
    }

    /**
     * Used to process a status information as needed by the client object (i.e.
     * errors or success).
     *
     * @param statusCode the status code
     * @param statusMessage additional descriptive message
     *
     * @see sync4j.framework.core.StatusCode for valid status codes.
     */
    public SyncResponse processStatusCode(int statusCode, String info){
        if (statusCode != StatusCode.OK) {
            sessionHandler.abort(statusCode);
        }
        return null;
    }

    // --------------------------------------------------------- private methods

    /**
     * Processes the input SyncML message after convertion to objects. See class
     * description for more information.
     *
     * @return the response message
     *
     * @throws ServerException in case of errors
     */
    private SyncML processInputMessage(final SyncML syncMLIn) throws ServerException {
        try {
            try {
                sessionHandler.setSizeCalculator(
                    getSizeCalculator()
                );
                return sessionHandler.processMessage(syncMLIn, mpc);
            } catch (InvalidCredentialsException e) {
                return sessionHandler.processError(syncMLIn, e);
            } catch (ServerException e) {
                return sessionHandler.processError(syncMLIn, e);
            } catch (ProtocolException e) {
                return sessionHandler.processError(syncMLIn,
                                       new BadRequestException(e.getMessage()));
            }

        } catch (Sync4jException e1) {
            //
            // This can be due only to processError
            //
            throw new ServerException(e1);
        }
    }

    /**
     * Converts the given SyncML message into a <i>SyncML</i> object.
     *
     * @param msg the SyncML message
     *
     * @return the corresponding SyncML message
     *
     * @throws ServerException in case of translating errors
     */
    private SyncML convert(final String msg) throws ServerException {
        try {
            IBindingFactory f = BindingDirectory.getFactory(SyncML.class);
            IUnmarshallingContext c = f.createUnmarshallingContext();

            Object syncML = c.unmarshalDocument(new StringReader(msg));

            return (SyncML)syncML;
        } catch(JiBXException e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error unmarshalling message:" + e.getMessage());
            }
            log.throwing(getClass().getName(), "convert", e);
            throw new ServerException(e);
        }
    }

    /**
     * Converts a <i>SyncML</i> object into the corresponding SyncML message (as
     * String).
     *
     * @param msg the SyncML object
     *
     * @return the corresponding SyncML message
     *
     * @throws ServerException in case of translation errors
     */
    private byte[] convert(final SyncML msg, String charset)
    throws ServerException {
        try {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Creating response with charset: " + charset);
            }

            ByteArrayOutputStream bout = new ByteArrayOutputStream();
            IBindingFactory f = BindingDirectory.getFactory(SyncML.class);
            IMarshallingContext c = f.createMarshallingContext();
            c.setIndent(0);
            c.marshalDocument(msg, charset, null, bout);

            //
            // Turn the output message in the canonical form and return it
            //
            return syncMLCanonizer.canonizeOutput(bout.toString(charset)).getBytes(charset);

        } catch(Exception e) {
            if (log.isLoggable(Level.SEVERE)) {
                log.severe("Error in converting the message:" + e.getMessage());
            }
            log.throwing(getClass().getName(), "convert", e);

            throw new ServerException(e);
        }
    }

    /**
     * Set the RespURI in the given message
     *
     * @param msg the message into wich set the RespURI
     * @param sessionId the session id
     */
    private void setRespURI(SyncML msg, final String sessionId) {
        msg.getSyncHdr().setRespURI(
            config.getServerConfig().getEngineConfiguration().getServerURI() +
            '?' +
            PARAM_SESSION_ID +
            '=' +
            sessionId
        );
    }

    /**
     * Checks if the RespURI element is going to be bigger than the space we
     * reserved for it. In this case we trace a worning in the log. Note that
     * not necessarily a RespURI bigger than the save space will result in a
     * message bigger then the maxmsgsize, therefore a warning looks the more
     * appropriate level at which logging it.
     *
     * @param uri the resp URI
     */
    private void checkRespURISize(String uri) {
        //
        // The worse case is using XML: <RespURI></RespURI> -> 20 chars
        //
        final long s = uri.length() + 20;
        final long h = getSizeCalculator().getRespURIOverhead();

        if ((log.isLoggable(Level.WARNING)) && (h < s)) {
            log.warning( "The RespURI element size ("
                       + s
                       + ") exeeds the reserved hoverhead ("
                       + h
                       + ')'
                       );
        }
    }

    /**
     * Returns the MessageSizeCalculator to be used accordingly to the message
     * content type (retrieved by the processing context)
     *
     * @return
     */
    private MessageSizeCalculator getSizeCalculator() {
        Map headers = (Map)mpc.getRequestProperty(mpc.PROPERTY_REQUEST_HEADERS);

        String contentType = (String)headers.get(HEADER_CONTENT_TYPE);

        MessageSizeCalculator calculator = null;

        if ((contentType != null) && Constants.MIMETYPE_SYNCMLDS_WBXML.equals(contentType)) {
            calculator = new WBXMLSizeCalculator();
        } else {
            calculator = new XMLSizeCalculator();
        }

        return calculator;
    }

    // ------------------------------------------------------------ SyncResponse

    private static class Sync4jResponse implements SyncResponse {
        private byte[] msg;
        private boolean completed;

        /**
         * Create a new Sync4jResponse.
         *
         * @param msg the marshalled response
         * @param syncML the SyncML object
         * @param resultMimeType the mime type
         *
         */
        private Sync4jResponse(final byte[]  msg,
                               final SyncML  syncML) {
            this.msg            = msg;
            this.completed = (syncML.isLastMessage() &&
                             ProtocolUtil.noMoreResponse(syncML) &&
                             (ProtocolUtil.getStatusChal(syncML) == null));
        }

        public byte[] getMessage() {
            return this.msg;
        }

        /**
         * Is this message the last message allowed for the current session?
         *
         * @return true if yes, false otherwise
         */
        public boolean isCompleted() {
            return completed;
        }
    }
}
