/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package sync4j.server.store;

import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

import java.sql.*;

import sync4j.framework.tools.DBTools;

import sync4j.framework.server.ClientMapping;
import sync4j.framework.server.store.Clause;
import sync4j.framework.server.store.PersistentStore;
import sync4j.framework.server.store.BasePersistentStore;
import sync4j.framework.server.store.NotFoundException;
import sync4j.framework.server.store.PersistentStoreException;

/**
 * This is the store for information regarding the synchronization process and
 * status. It persists the following classes:
 *
 * <ul>
 * <li>sync4j.framework.server.ClientMapping</li>
 * </ul>
 *
 * This <i>PersistentStore</i> is configured with the following map key:
 * <ul>
 *  <li>jndiDataSourceName</li>
 * </ul>
 *
 * @author Stefano Fornari @ Funambol
 * @author Harrie Hazewinkel
 * @version $Id: ClientMappingPersistentStore.java,v 1.6 2005/04/20 23:19:13 stefano_fornari Exp $
 */
public class ClientMappingPersistentStore
extends    BasePersistentStore
implements PersistentStore, java.io.Serializable {

    // --------------------------------------------------------------- Constants

    // -------------------------------------------------------------- Properties
    // --------------------------------------------- configurable SQL properties
    private String sqlInsertClientMapping =
    "insert into sync4j_client_mapping (principal, sync_source, guid, luid) values(?, ?, ?, ?)";
    private String sqlDeleteClientMapping =
    "delete from sync4j_client_mapping where principal=? and sync_source=? and luid=?";
    private String sqlUpdateClientMapping =
    "update sync4j_client_mapping set luid=? where principal=? and sync_source=? and guid=?";
    private String sqlSelectClientMapping =
    "select luid,guid from sync4j_client_mapping where principal=? and sync_source=?";

    /** Getter for property sqlInsertClientMapping.
     * @return Value of property sqlInsertClientMapping.
     *
     */
    public String getSqlInsertClientMapping() {
        return sqlInsertClientMapping;
    }

    /** Setter for property sqlInsertClientMapping.
     * @param sqlInsertClientMapping New value of property sqlInsertClientMapping.
     *
     */
    public void setSqlInsertClientMapping(String sqlInsertClientMapping) {
        this.sqlInsertClientMapping = sqlInsertClientMapping;
    }

    /** Getter for property sqlDeleteClientMapping.
     * @return Value of property sqlDeleteClientMapping.
     *
     */
    public String getSqlDeleteClientMapping() {
        return sqlDeleteClientMapping;
    }

    /** Setter for property sqlDeleteClientMapping.
     * @param sqlDeleteClientMapping New value of property sqlDeleteClientMapping.
     *
     */
    public void setSqlDeleteClientMapping(String sqlDeleteClientMapping) {
        this.sqlDeleteClientMapping = sqlDeleteClientMapping;
    }

    /** Getter for property sqlUpdateClientMapping.
     * @return Value of property sqlUpdateClientMapping.
     *
     */
    public String getSqlUpdateClientMapping() {
        return sqlUpdateClientMapping;
    }

    /** Setter for property sqlUpdateClientMapping.
     * @param sqlUpdateClientMapping New value of property sqlUpdateClientMapping.
     *
     */
    public void setSqlUpdateClientMapping(String sqlUpdateClientMapping) {
        this.sqlUpdateClientMapping = sqlUpdateClientMapping;
    }

    /** Getter for property sqlSelectClientMapping.
     * @return Value of property sqlSelectClientMapping.
     *
     */
    public String getSqlSelectClientMapping() {
        return sqlSelectClientMapping;
    }

    /** Setter for property sqlSelectClientMapping.
     * @param sqlSelectClientMapping New value of property sqlSelectClientMapping.
     *
     */
    public void setSqlSelectClientMapping(String sqlSelectClientMapping) {
        this.sqlSelectClientMapping = sqlSelectClientMapping;
    }

    // ------------------------------------------------------------ Private data

    // ------------------------------------------------------------ Constructors

    // ---------------------------------------------------------- Public methods

    public boolean store(Object o)
    throws PersistentStoreException {
        if (o instanceof ClientMapping) {
            ClientMapping clientMapping = (ClientMapping) o;

            Connection conn = null;
            PreparedStatement stmt = null, stmtIns = null;

            long principal = -1;
            String dbURI     = clientMapping.getDbURI();
            
            principal = Long.parseLong(clientMapping.getPrincipal().getId());

            assert ((principal >= 0) && (dbURI != null));

            try {
                conn = dataSource.getConnection();

                if (clientMapping.isDeleted()) {
                    stmt = conn.prepareStatement(sqlDeleteClientMapping);
                    stmt.setLong(1, principal);
                    stmt.setString(2, dbURI    );
                    String[] luids = clientMapping.getDeletedLuids();
                    for (int i=0; i < luids.length; ++i) {
                        stmt.setString(3, luids[i]);
                        stmt.executeUpdate();
                    }
                    stmt.close(); stmt = null;
                }

                if (clientMapping.isModified()) {
                    stmt = conn.prepareStatement(sqlUpdateClientMapping);
                    stmt.setLong(2, principal);
                    stmt.setString(3, dbURI    );
                    String[] luids = clientMapping.getModifiedLuids();

                    String guid = null;
                    for (int i=0; i < luids.length; ++i) {
                        int n = -1;

                        guid = clientMapping.getMappedValueForLuid(luids[i]);
                        stmt.setString(1, luids[i]);
                        stmt.setString(4, guid    );
                        n = stmt.executeUpdate();

                        if (n == 0) {
                            //
                            // insert new mapping
                            //
                            stmtIns = conn.prepareStatement(sqlInsertClientMapping);
                            stmtIns.setLong(1, principal);
                            stmtIns.setString(2, dbURI    );
                            stmtIns.setString(3, guid);
                            stmtIns.setString(4, luids[i]);
                            stmtIns.executeUpdate();
                            stmtIns.close(); stmtIns = null;
                        }
                    }
                    stmt.close(); stmt = null;
                }
            } catch (SQLException e) {
                throw new PersistentStoreException("Error storing client mapping", e);
            } finally {
                DBTools.close(conn, stmt, null);
            }
            return true;
        }
        return false;
    }

    public boolean read(Object o)
    throws PersistentStoreException {
        if (o instanceof ClientMapping) {
            ClientMapping clientMapping = (ClientMapping) o;

            Connection conn = null;
            PreparedStatement stmt = null;
            ResultSet rs = null;

            try {
                conn = dataSource.getConnection();

                stmt = conn.prepareStatement(sqlSelectClientMapping);
                stmt.setLong(1, Long.parseLong(clientMapping.getPrincipal().getId()));
                stmt.setString(2, clientMapping.getDbURI()            );
                rs = stmt.executeQuery();

                HashMap mapping = new HashMap();
                while (rs.next()) {
                    mapping.put(rs.getString("guid"), rs.getString("luid"));
                }
                clientMapping.initializeFromMapping(mapping);
            } catch (SQLException e) {
                throw new PersistentStoreException("Error reading mapping", e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
            return true;
        }
        return false;
    }

    /** Read all objects stored the persistent media.
     *
     * @param objClass the object class handled by the persistent store
     *
     * @return an array containing the objects read. If no objects are found an
     *         empty array is returned. If the persistent store has not
     *         processed the quest, null is returned.
     *
     * @throws PersistentStoreException
     *
     */
    public Object[] read(Class objClass) throws PersistentStoreException {
        //
        // TO DO (not used yet)
        //
        return null;
    }

    public boolean delete(Object o) throws PersistentStoreException
    {
        return false;
    }

    public Object[] read(Object o, Clause clause) throws PersistentStoreException
    {
        return null;
    }

    public int count(Object o, Clause clause) throws PersistentStoreException
    {
        return -1;
    }

    public boolean store(Object o, String operation) throws PersistentStoreException
    {
        return false;
    }

}
