/**
 * Copyright (C) 2003-2005 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


package sync4j.transport.http.server;

import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import java.util.logging.Level;

import javax.naming.*;
import javax.servlet.*;
import javax.servlet.http.*;

import sync4j.framework.core.*;
import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.server.*;
import sync4j.framework.tools.Base64;
import sync4j.framework.transport.http.SyncHolder;
import sync4j.framework.transport.http.SyncHolderCache;
import sync4j.framework.protocol.ProtocolException;

import sync4j.server.config.Configuration;

import org.apache.commons.lang.StringUtils;

import org.jibx.runtime.*;
import org.jibx.runtime.impl.*;

/**
 *  Receives the HTTP request and does session management.
 *
 *  <b>Session Management</b>
 *  <p>
 *  If the url contains a parameter {PARAM_SESSION_ID}={SESSION_ID}, the session
 *  id is used to lookup in the <i>handlerCache</i> if there is already a
 *  <i>SyncHolder</i> associated with the given session id it is used to process
 *  the incoming request. Otherwise (either if the session parameter is not
 *  specified or not cached handler is found), a new session id is created and a
 *  new <i>SyncHolder</i> object is instantiated and stored in the cache.<br>
 *  The session id is created as followed:
 *  <p>
 *  <ol>
 *    <li> the first four bytes contains the IP address of the remote client
 *    <li> a dash ('-') is appended
 *    <li> the creation timestamp is appended
 *    <li> the resulting string is encoded base 64
 *  </ol>
 *  Note that no session expiration is handled by this class. That is delegated
 *  to the <i>SyncHolderCache</i> object.
 *
 *  @author Stefano Fornari @ Funambol
 *
 *  @version $Id: Sync4jServlet.java,v 1.20 2005/05/27 12:04:53 nichele Exp $
 *
 */
public final class Sync4jServlet
extends HttpServlet
implements Constants {

    // --------------------------------------------------------------- Constants

    // ------------------------------------------------------------ Private data

    private Logger log = Sync4jLogger.getLogger(LOG_NAME);

    private static String          syncHolderClass = null;
    private static SyncHolderCache holderCache     = null;

    //
    // To check if the engine was initialized
    //
    private boolean initialized = false;
    private static final String PATH_STATUS = "/sync4j/sync/status";

    private static final int SIZE_INPUT_BUFFER = 4096;
    // ---------------------------------------------------------- Public methods

    public void init() throws ServletException {
        initialized = false;
        String value;
        long timeToLive = 0;

        if (log.isLoggable(Level.INFO)) {
            System.getProperties().list(System.out);
        }

        ServletConfig config = getServletConfig();

        value = config.getInitParameter(PARAM_SYNCHOLDER_CLASS);

        if (StringUtils.isEmpty(value)) {
            String msg = "The servlet configuration parameter "
            + PARAM_SYNCHOLDER_CLASS
            + " cannot be empty ("
            + value
            + ")"
            ;
            log(msg);
            throw new ServletException(msg);
        }
        syncHolderClass = value;

        value = config.getInitParameter(PARAM_SESSION_TTL);
        if (!StringUtils.isEmpty(value)) {
            try {
                timeToLive = Long.parseLong(value);
            } catch (NumberFormatException e) {
                String msg = "The servlet configuration parameter "
                + PARAM_SESSION_TTL
                + " is not an integer number ("
                + value
                + ")"
                ;
                log.severe(msg);
                throw new ServletException(msg);
            }
        }

        holderCache = new SyncHolderCache(timeToLive);

        //
        // Now we are ready to bootstrap the Sync4j components
        //
        bootstrap();

        //
        // Engine initialized!
        //
        initialized = true;
    }


    public void doPost(final HttpServletRequest  httpRequest ,
                       final HttpServletResponse httpResponse)
    throws ServletException, IOException {

        if (log.isLoggable(Level.FINEST)) {
            showHeaders(httpRequest);
        }

        if (log.isLoggable(Level.INFO)) {
            StringBuffer sb = new StringBuffer("Handling incoming request ");
            sb.append(httpRequest.getContextPath())
              .append(httpRequest.getServletPath());
            if (httpRequest.getPathInfo() != null) {
                sb.append(httpRequest.getPathInfo());
            }

            if (httpRequest.getQueryString() != null) {
                sb.append('?').append(httpRequest.getQueryString());
            }
            sb.append('.');

            log.info(sb.toString());
        }

        final String contentType = httpRequest.getContentType().split(";")[0];
        final int contentLength  = httpRequest.getContentLength();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("contentType: " + contentType);
            log.finest("contentLength: " + contentLength);
        }


        Map params  = getRequestParameters(httpRequest);
        Map headers = getRequestHeaders   (httpRequest);

        byte[] requestData = null;
        InputStream in = httpRequest.getInputStream();

        try{
            requestData = readContent(in);
        } catch (Exception e) {
            handleError(httpRequest, httpResponse, "Error reading the request", e);
            return;
        } finally {

            try {
                if (in != null) {
                    in.close();
                }
            } catch (Exception ex) {
                handleError(httpRequest, httpResponse, ex.getClass().getName(), ex);
                return;
            }
            in = null;
        }

        //
        // If the session id is not specified in the URL, a new remote object
        // will be created. Otherwise the session id specifies which remote
        // object shall handles the request.
        //
        SyncHolder holder = null;
        try {
            holder = createHolder(httpRequest);
        } catch (Exception e) {
            handleError(httpRequest, httpResponse, "Error creating SyncBean", e);
            return;
        }

        SyncResponse resp = null;
        try {
            if (sync4j.framework.core.Constants.MIMETYPE_SYNCMLDS_WBXML.equals(contentType)) {
                resp = holder.processWBXMLMessage(requestData, params, headers);
            } else if (sync4j.framework.core.Constants.MIMETYPE_SYNCMLDS_XML.equals(contentType)) {
                resp = holder.processXMLMessage(requestData, params, headers);
            } else {
                throw new ProtocolException( "Mime type "
                                           + contentType
                                           + " nor supported or unknown" );
            }
        } catch (Exception e) {
            Throwable cause = e.getCause();

            if (  (cause != null)
            && (  (cause instanceof ProtocolException)
            || (cause instanceof Sync4jException  )
            )
            ) {
                handleError(httpRequest, httpResponse, "Protocol error", cause);
                return;
            } else {
                throw new ServletException(e);
            }
        }

        OutputStream out = null;
        try {
            out = httpResponse.getOutputStream();
            httpResponse.setContentType(contentType);

            byte[] response = resp.getMessage();

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Outgoing message length: " + response.length);
            }

            httpResponse.setContentLength(response.length);
            out.write(response);
            out.flush();

        } finally {
            if (out != null) {
                out.close();
            }
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Finally");
            }
        }

        //
        // If the message completed the SyncML communication, the holder
        // must be closed and discarded.
        //
        if (resp.isCompleted()) {
            releaseHolder(holder);
        }
    }

    public void doGet(final HttpServletRequest  httpRequest ,
                      final HttpServletResponse httpResponse)
    throws ServletException, IOException {

        String requestURI = httpRequest.getRequestURI();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("requestURI: " + requestURI);
        }
        if (requestURI.equals(PATH_STATUS)) {
            if (initialized) {
                //
                //The request sent by the client was successful
                //
                httpResponse.setStatus(httpResponse.SC_OK);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Request succeeded normally");
                }
            } else {
                //
                //The request was unsuccessful to the server being down or overloaded.
                //
                httpResponse.setStatus(httpResponse.SC_SERVICE_UNAVAILABLE);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("The HTTP server is unable to handle the request");
                }
            }
        }
    }

    // ------------------------------------------------------- Protected methods
    /**
     * Reads the request body without to known the length of the data.
     *
     * @param in the request InputStream
     *
     * @throws IOException in case of errors
     */
    protected byte[] readContent(final InputStream in) throws IOException {
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        byte[] buf = new byte[SIZE_INPUT_BUFFER];

        int c = 0;
        int b = 0;
        while ((c < buf.length) && (b = in.read(buf, c, buf.length-c)) >= 0) {
            c+=b;
            if (c == SIZE_INPUT_BUFFER) {
                bout.write(buf);
                buf = new byte[SIZE_INPUT_BUFFER];
                c = 0;
            }
        }
        if (c != 0) {
            bout.write(buf, 0, c);
        }
        return bout.toByteArray();
    }

    // --------------------------------------------------------- Private methods
    /**
     * Factory method for <i>SyncHolder</i> objects. If the session id is
     * passed as a CGI parameter, it is transformed in a EJB handle. If the
     * handle is not valid, it is considered expired. If the session id is not
     * specified, a new EJB is created
     *
     * @param request the associated HTTP request object
     *
     * @return a new <i>SyncHolder</i>
     *
     */
    private SyncHolder createHolder(HttpServletRequest request)
    throws Exception {
        String sessionId = request.getParameter(PARAM_SESSION_ID);

        SyncHolder holder = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("cache: " + holderCache);
        }

        if (!StringUtils.isEmpty(sessionId)) {
            holder = (SyncHolder)holderCache.get(sessionId);
        }

        if (holder == null) {
            holder = (SyncHolder)getClass().forName(syncHolderClass).newInstance();

            sessionId = createSessionId(request);
            holder.setSessionId(sessionId);
            holderCache.put(holder);
        }

        return holder;
    }

    /**
     * Closes the given <i>SyncHolder</i> and removes it from the cache.
     *
     * @param holder the holder to releases
     */
    private void releaseHolder(SyncHolder holder) {
        try {
            holder.close();
        } catch (Exception e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "releaseHolder", e.getCause());
        }
        holderCache.remove(holder.getSessionId());
    }

    /**
     * Handles errors conditions returning an appropriate content to the client.
     *
     * @param request the request object
     * @param response the response object
     * @msg   a desctiptive message
     * @t     a throwable object
     *
     */
    private void handleError(final HttpServletRequest  request ,
                             final HttpServletResponse response,
                             String                    msg     ,
                             final Throwable           t       ) {

        if (msg == null) {
            msg = "";
        }

        if (t == null) {
            log.severe(msg);
        } else {
            log.severe(msg);
            log.throwing(getClass().getName(), "unknown", t);
        }
        try {
            response.sendError(response.SC_BAD_REQUEST, msg);
        } catch (IOException e) {
            log.severe(e.getMessage());
            log.throwing(getClass().getName(), "unknown", t);
        }
    }

    /**
     * Creates the session id (see the class description for details).
     *
     * @param request the HTTP request object
     *
     * @return a newly created session id
     */
    private String createSessionId(HttpServletRequest request) {
        String clientIP  = request.getRemoteAddr()   ;
        long   timestamp = System.currentTimeMillis();

        StringBuffer sb = new StringBuffer();

        byte[] ip = null;

        try {
            ip = InetAddress.getByName(clientIP).getAddress();
        } catch (UnknownHostException e) {
            ip = new byte[] {0, 0, 0, 0};
        }

        sb.append(ip).append('-').append(timestamp);

        String stringSessionId = new String(Base64.encode(sb.toString().getBytes()));

        //
        // strips out tailing '='
        //
        int i = stringSessionId.lastIndexOf('=');

        return ((i>0) ? stringSessionId.substring(0, i-1) : stringSessionId);
    }

    /**
     * Extracts the request parameters and returns them in a <i>Map</i>
     *
     * @param request the request
     */
    private Map getRequestParameters(HttpServletRequest request) {
        Map params = new HashMap();

        String paramName = null;
        for (Enumeration e=request.getParameterNames(); e.hasMoreElements(); ) {
            paramName = (String)e.nextElement();

            params.put(paramName, request.getParameter(paramName));
        }

        return params;
    }

    /**
     * Extracts the request headers and returns them in a <i>Map</i>
     *
     * @param request the request
     */
    private Map getRequestHeaders(HttpServletRequest request) {
        Map headers = new HashMap();

        String headerName = null;
        for (Enumeration e=request.getHeaderNames(); e.hasMoreElements(); ) {
            headerName = (String)e.nextElement();

            headers.put(headerName.toLowerCase(), request.getHeader(headerName));
        }

        return headers;
    }

    /**
     * The same as <i>handleError(reqest, response, msg, t, null)</i>.
     *
     * @param request the request object
     * @param response the response object
     * @msg   a desctiptive message
     * @t     a throwable object
     *
     */
    private void handleError(final HttpServletRequest   request,
                             final HttpServletResponse response,
                             final String                   msg) {
        this.handleError(request, response, msg, null);
    }

    /**
     * Bootstraps the Sync4j server components and logs the server startup
     * event on both the console and the logging system.
     */
    private void bootstrap() {
        Configuration c = null;

        c = Configuration.getConfiguration();

        String msg1 = "================================================================================";
        String sp   = "\n\n";
        String msg2 = "SyncServer v. "
                    + c.getServerConfig().getServerInfo().getSwV()
                    + " engine started.\nConfiguration object found."
                    ;

        //
        // NOTE: this code uses the stdout by choice, so that it will always
        // be displayed, regardless how logging is configured.
        //
        System.out.println(msg1);
        System.out.println(sp);
        System.out.println(msg2);
        System.out.println(sp);
        System.getProperties().list(System.out);
        System.out.println(sp);
        System.out.println(msg1);

        if (log.isLoggable(Level.INFO)) {
            log.info(msg1+sp);
            log.info(msg2+sp);
            log.info(System.getProperties().toString()+sp);
            log.info(msg1);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Engine configuration:");
            log.finest(" - store: "
                      + c.getServerConfig().getEngineConfiguration().getStore()
                      + " (" + c.getStore() + ")"
            );
            log.finest(" - officer: "
                      + c.getServerConfig().getEngineConfiguration().getOfficer()
                      + " (" + c.getOfficer() + ")"
            );
            log.finest(" - strategy: "
                      + c.getServerConfig().getEngineConfiguration().getStrategy()
                      + " (" + c.getStrategy() + ")"
            );

            log.finest("Default encoding: "
                       + (new OutputStreamWriter(new ByteArrayOutputStream())).
                      getEncoding()
            );

        }
    }

    /**
     * Log the headers of the request
     * @param request HttpServletRequest
     */
    private void showHeaders(HttpServletRequest request) {
        Enumeration enumHeaders = request.getHeaderNames();
        String headerName = null;
        log.finest("Http header: ");
        while (enumHeaders.hasMoreElements()) {
            headerName = (String)enumHeaders.nextElement();
            log.finest(headerName + ": " + request.getHeader(headerName));
        }
    }
}
